import pytest

import piglr


def test_init():
    game = piglr.Game(2, 1, 6, 100)
    game.reset()

    assert game.players == 2
    assert game.num_dice == 1
    assert game.dn == 6
    assert game.target == 100


def test_init_errors():
    with pytest.raises(piglr.error.IllegalParameterError):
        # Check for fewer than 2 players
        game = piglr.Game(1, 1, 6, 100)
        game.reset()
    with pytest.raises(piglr.error.IllegalParameterError):
        # Check for fewer than 1 die
        game = piglr.Game(2, 0, 6, 100)
        game.reset()
    with pytest.raises(piglr.error.IllegalParameterError):
        # Check for die with fewer than 2 sides
        game = piglr.Game(2, 1, 1, 100)
        game.reset()
    with pytest.raises(piglr.error.IllegalParameterError):
        # Check for score < 1
        game = piglr.Game(2, 1, 6, 0)
        game.reset()


def test_winner():
    game = piglr.Game(2, 1, 6, 100)
    game.reset()

    assert game.target == 100

    score = game._state.score
    assert score[0] == 0
    assert score[1] == 0

    winner = game._check_for_winner()
    assert winner is None

    game._state.score[0] = 100
    assert game.target == game._state.score[0]

    winner = game._check_for_winner()
    assert winner is not None


def test_bank():
    game = piglr.Game(2, 1, 6, 100)
    game.reset()

    score = game._state.score
    assert score[0] == 0
    assert score[1] == 0

    assert game._state.bank == 0
    game._state.turn = 1
    game._state.rolls = 1
    game._state.bank = 1

    assert game._state.bank == 1

    game._increment_turn(True)

    assert game._state.bank == 0




"""Piglr Game module."""

import random
from typing import Union

from piglr.state import State, as_dict
from piglr.error import IllegalParameterError


class Game:
    """Game class used to interact with a the pig environment.

    Example usage:
    
    >>> import json
    >>> import random
    >>> import piglr
    >>> game = piglr.Game()
    >>> obs, winner = game.reset()
    >>> while winner is None:
    >>>     roll = random.randint(0, 1)
    >>>     if roll:
    >>>         obs, winner = game.roll()
    >>>     else:
    >>>         obs, winner = game.bank()
    >>> print(json.dumps(obs, indent=4))
    >>> print(f'Winner: {winner}')

    """
    def __init__(self, players: int = 2, num_dice: int = 1, dn: int = 6, target: int = 100):
        """`Game` init method.

        Args:
            players: Number of players. (Legal range 2-inf)
            num_dice: Number of dice. (Legal range 1-inf)
            dn: Dice sidedness (Think D&D, d6, d8, d20, etc.). (Legal
                range 1-inf)
            target: Target score for the game. (Legal range 1-inf)

        Raises:
            TypeError: TypeError is raised whenever one of the passed
                parameters is not of type int.
            piglr.error.IllegalParameterError: is raised whenever a
                parameter is passed an illegal value. For legal values
                see args section of init method.

        """
        for param in [players, num_dice, dn, target]:
            if not isinstance(param, int):
                param_name = f'{param}='.split('=')[0]
                raise TypeError(f'Param {param_name} must be of type int')

        if players < 2:
            raise IllegalParameterError('There must be at least 2 players')

        if num_dice < 1:
            raise IllegalParameterError('There must be at least 1 dice')

        if dn < 2:
            raise IllegalParameterError('The dice must be at least 2 sided')

        if target < 1:
            raise IllegalParameterError('The target must be positive')

        self.players = players
        self.num_dice = num_dice
        self.dn = dn
        self.target = target

    @property
    def state(self) -> dict:
        """State property.

        Returns:
            Instance of :obj:`piglr.state.State` converted to a
            dictionary.

        """
        return as_dict(self._state)

    def _increment_turn(self, bank: bool) -> None:
        if bank:
            self._state.score[self._state.turn] += self._state.bank

        self._state.rolls = 0
        self._state.bank = 0

        if self._state.players - 1 == self._state.turn:
            self._state.turn = 0
        else:
            self._state.turn += 1

    def _check_for_winner(self) -> Union[int, None]:
        if max(self._state.score.values()) >= self._state.target:
            return max(self._state.score, key=self._state.score.get)
        return None

    def reset(self) -> Union[dict, None]:
        """Reset method.

        Resets the game state.

        Returns:
            - State property :attr:`piglr.game.Game.state`.
            - :obj:`None`.

        """
        self._state = State(
            players=self.players,
            num_dice=self.num_dice,
            dn=self.dn,
            score={i:0 for i in range(self.players)},
            target=self.target
        )

        return self.state, None

    def roll(self) -> Union[dict, None]:
        """Roll dice.

        Rolls dice for current player.

        Returns:
            - State property :attr:`piglr.game.Game.state`.
            - :obj:`None`.

        """
        for _ in range(self._state.num_dice):
            d = random.randint(1, self._state.dn)
            if d == 1:
                self._increment_turn(bank=False)
                return self.state, None

            else:
                self._state.rolls += 1
                self._state.bank += d

        return self.state, None
    
    def bank(self) -> Union[dict, Union[int, None]]:
        """Scores bank.

        Scores bank for current player.        

        Returns:
            - State property :attr:`piglr.game.Game.state`.
            - :obj:`int` if the game has been won else :obj:`None`.

        """
        self._increment_turn(bank=True)

        winner = self._check_for_winner()

        return self.state, winner



import setuptools
import sys
import os


sys.path.insert(0, os.path.join(os.path.dirname(__file__), 'piglr'))
from version import VERSION as __version__

with open('README.md', 'r') as f:
    long_description = f.read()

with open('LICENSE', 'r') as f:
    license = f.read()

setuptools.setup(
    name='piglr',
    version=__version__,
    author='Christian Lang',
    author_email='me@christianlang.io',
    description='A Pig Reinforcement Learning Environment.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/christianwlang/piglr',
    download_url='https://pypi.org/project/piglr/',
    project_urls={
        'Documentation': 'https://piglr.readthedocs.io/en/latest/',
        'Source': 'https://github.com/christianwlang/piglr',
        'Tracker': 'https://github.com/christianwlang/piglr/issues'
    },
    packages=setuptools.find_packages(),
    include_package_data=True,
    license=license,
    extras_require={
        'testing': ['pytest', 'pytest-cov', 'flake8'],
        'docs': ['sphinx']
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent'
    ]
)



"""Piglr gamestate module."""

from dataclasses import dataclass, asdict
from typing import Dict


@dataclass
class State:
    """Piglr state dataclass."""
    players: int
    """Number of players."""
    num_dice: int
    """Number of dice."""
    dn: int
    """Dice sidedness (Think D&D, d6, d8, d20, etc.)."""
    score: Dict[int, int]
    """Current game score."""
    target: int
    """Target score for the game."""
    turn: int = 0
    """Active player (who's turn is it)."""
    rolls: int = 0
    """Number of time's active player has rolled."""
    bank: int = 0
    """Active player's bank."""


def as_dict(state: State) -> dict:
    """See :func:`dataclasses.asdict`."""
    return asdict(state)
